@@ -1,5 +1,6 @@
 /*
  * Copyright 2013,2014,2015,2016,2021 Vladimir Ermakov.
+ * Copyright 2022 Dr.-Ing. Amilcar do Carmo Lucas, IAV GmbH.
  *
  * This file is part of the mavros package and subject to the license terms
  * in the top-level LICENSE file of the mavros repository.
@@ -14,10 +15,12 @@
  * @{
  */
 
+#include <algorithm>
 #include <string>
 #include <vector>
 #include <unordered_map>
 #include <memory>
+#include <utility>
 
 #include "rcpputils/asserts.hpp"
 #include "mavros/mavros_uas.hpp"
@@ -45,14 +48,17 @@ using mavlink::minimal::MAV_AUTOPILOT;
 using mavlink::minimal::MAV_STATE;
 using utils::enum_value;
 using BatteryMsg = sensor_msgs::msg::BatteryState;
+using DiagnosticStatus = diagnostic_msgs::msg::DiagnosticStatus;
 
 using namespace std::placeholders;      // NOLINT
 using namespace std::chrono_literals;   // NOLINT
 
+#define MAX_NR_BATTERY_STATUS 10
+
 /**
  * Heartbeat status publisher
  *
- * Based on diagnistic_updater::FrequencyStatus
+ * Based on diagnostic_updater::FrequencyStatus
  */
 class HeartbeatStatus : public diagnostic_updater::DiagnosticTask
 {
@@ -113,13 +119,13 @@ public:
     hist_indx_ = (hist_indx_ + 1) % window_size_;
 
     if (events == 0) {
-      stat.summary(2, "No events recorded.");
+      stat.summary(DiagnosticStatus::ERROR, "No events recorded.");
     } else if (freq < min_freq_ * (1 - tolerance_)) {
-      stat.summary(1, "Frequency too low.");
+      stat.summary(DiagnosticStatus::WARN, "Frequency too low.");
     } else if (freq > max_freq_ * (1 + tolerance_)) {
-      stat.summary(1, "Frequency too high.");
+      stat.summary(DiagnosticStatus::WARN, "Frequency too high.");
     } else {
-      stat.summary(0, "Normal");
+      stat.summary(DiagnosticStatus::OK, "Normal");
     }
 
     stat.addf("Heartbeats since startup", "%d", count_);
@@ -173,9 +179,9 @@ public:
     if ((last_st.onboard_control_sensors_health & last_st.onboard_control_sensors_enabled) !=
       last_st.onboard_control_sensors_enabled)
     {
-      stat.summary(2, "Sensor health");
+      stat.summary(DiagnosticStatus::ERROR, "Sensor health");
     } else {
-      stat.summary(0, "Normal");
+      stat.summary(DiagnosticStatus::OK, "Normal");
     }
 
     stat.addf("Sensor present", "0x%08X", last_st.onboard_control_sensors_present);
@@ -270,12 +276,34 @@ class BatteryStatusDiag : public diagnostic_updater::DiagnosticTask
 public:
   explicit BatteryStatusDiag(const std::string & name)
   : diagnostic_updater::DiagnosticTask(name),
-    voltage(-1.0),
-    current(0.0),
-    remaining(0.0),
-    min_voltage(6)
+    voltage(-1.0f),
+    current(0.0f),
+    remaining(0.0f),
+    min_voltage(6.0f)
   {}
 
+  // Move constructor, required to dynamically create an array of instances of this class
+  // because it contains an unique mutex object
+  BatteryStatusDiag(BatteryStatusDiag && other) noexcept
+  : diagnostic_updater::DiagnosticTask(""),
+    voltage(-1.0f),
+    current(0.0f),
+    remaining(0.0f),
+    min_voltage(6.0f)
+  {
+    *this = std::move(other);
+  }
+
+  // Move assignment operator, required to dynamically create an array of instances of this class
+  // because it contains an unique mutex object
+  BatteryStatusDiag & operator=(BatteryStatusDiag && other) noexcept
+  {
+    if (this != &other) {
+      *this = std::move(other);
+    }
+    return *this;
+  }
+
   void set_min_voltage(float volt)
   {
     std::lock_guard<std::mutex> lock(mutex);
@@ -290,21 +318,33 @@ public:
     remaining = rem;
   }
 
+  void setcell_v(const std::vector<float> voltages)
+  {
+    std::lock_guard<std::mutex> lock(mutex);
+    cell_voltage = voltages;
+  }
+
   void run(diagnostic_updater::DiagnosticStatusWrapper & stat)
   {
     std::lock_guard<std::mutex> lock(mutex);
 
-    if (voltage < 0) {
-      stat.summary(2, "No data");
+    if (voltage < 0.0f) {
+      stat.summary(DiagnosticStatus::ERROR, "No data");
     } else if (voltage < min_voltage) {
-      stat.summary(1, "Low voltage");
+      stat.summary(DiagnosticStatus::WARN, "Low voltage");
     } else {
-      stat.summary(0, "Normal");
+      stat.summary(DiagnosticStatus::OK, "Normal");
     }
 
     stat.addf("Voltage", "%.2f", voltage);
     stat.addf("Current", "%.1f", current);
-    stat.addf("Remaining", "%.1f", remaining * 100);
+    stat.addf("Remaining", "%.1f", remaining * 100.0f);
+    const int nr_cells = cell_voltage.size();
+    if (nr_cells > 1) {
+      for (int i = 1; i <= nr_cells; ++i) {
+        stat.addf(utils::format("Cell %u", i), "%.2f", cell_voltage[i - 1]);
+      }
+    }
   }
 
 private:
@@ -313,6 +353,7 @@ private:
   float current;
   float remaining;
   float min_voltage;
+  std::vector<float> cell_voltage;
 };
 
 
@@ -324,36 +365,51 @@ class MemInfo : public diagnostic_updater::DiagnosticTask
 public:
   explicit MemInfo(const std::string & name)
   : diagnostic_updater::DiagnosticTask(name),
-    freemem(-1),
-    brkval(0)
+    freemem(UINT32_MAX),
+    brkval(0),
+    last_rcd(0)
   {}
 
-  void set(uint16_t f, uint16_t b)
+  void set(uint32_t f, uint16_t b)
   {
     freemem = f;
     brkval = b;
+    last_rcd = clock.now().nanoseconds();
   }
 
   void run(diagnostic_updater::DiagnosticStatusWrapper & stat)
   {
-    ssize_t freemem_ = freemem;
+    // access atomic variables just once
+    size_t freemem_ = freemem;
     uint16_t brkval_ = brkval;
+    rclcpp::Time last_rcd_(last_rcd.load());
+    const rclcpp::Duration timeout(10s);              // seconds
 
-    if (freemem < 0) {
-      stat.summary(2, "No data");
-    } else if (freemem < 200) {
-      stat.summary(1, "Low mem");
+    // summarize the results
+    if (last_rcd_.nanoseconds() == 0) {
+      stat.summary(DiagnosticStatus::ERROR, "Not initialised");
+    } else if (clock.now() - last_rcd_ > timeout) {
+      stat.summary(
+        DiagnosticStatus::STALE,
+        "Not received for more than " + std::to_string(timeout.seconds()));
     } else {
-      stat.summary(0, "Normal");
+      if (freemem == UINT32_MAX) {
+        stat.summary(DiagnosticStatus::ERROR, "No data");
+      } else if (freemem < 200) {
+        stat.summary(DiagnosticStatus::WARN, "Low mem");
+      } else {
+        stat.summary(DiagnosticStatus::OK, "Normal");
+      }
     }
-
     stat.addf("Free memory (B)", "%zd", freemem_);
     stat.addf("Heap top", "0x%04X", brkval_);
   }
 
 private:
-  std::atomic<ssize_t> freemem;
+  rclcpp::Clock clock;
+  std::atomic<size_t> freemem;
   std::atomic<uint16_t> brkval;
+  std::atomic<uint64_t> last_rcd;
 };
 
 
@@ -367,40 +423,51 @@ public:
   : diagnostic_updater::DiagnosticTask(name),
     vcc(-1.0),
     i2cerr(0),
-    i2cerr_last(0)
+    i2cerr_last(0),
+    last_rcd(0)
   {}
 
   void set(uint16_t v, uint8_t e)
   {
     std::lock_guard<std::mutex> lock(mutex);
-    vcc = v / 1000.0;
+    vcc = v * 0.001f;
     i2cerr = e;
+    last_rcd = clock.now();
   }
 
   void run(diagnostic_updater::DiagnosticStatusWrapper & stat)
   {
     std::lock_guard<std::mutex> lock(mutex);
-
-    if (vcc < 0) {
-      stat.summary(2, "No data");
-    } else if (vcc < 4.5) {
-      stat.summary(1, "Low voltage");
-    } else if (i2cerr != i2cerr_last) {
-      i2cerr_last = i2cerr;
-      stat.summary(1, "New I2C error");
+    const rclcpp::Duration timeout(10s);               // seconds
+    if (last_rcd.nanoseconds() == 0) {
+      stat.summary(DiagnosticStatus::ERROR, "Not initialised");
+    } else if (clock.now() - last_rcd > timeout) {
+      stat.summary(
+        DiagnosticStatus::STALE, "Not received for more than " + std::to_string(
+          timeout.seconds()));
     } else {
-      stat.summary(0, "Normal");
+      if (vcc < 0) {
+        stat.summary(DiagnosticStatus::ERROR, "No data");
+      } else if (vcc < 4.5) {
+        stat.summary(DiagnosticStatus::WARN, "Low voltage");
+      } else if (i2cerr != i2cerr_last) {
+        i2cerr_last = i2cerr;
+        stat.summary(DiagnosticStatus::WARN, "New I2C error");
+      } else {
+        stat.summary(DiagnosticStatus::OK, "Normal");
+      }
     }
-
     stat.addf("Core voltage", "%f", vcc);
     stat.addf("I2C errors", "%zu", i2cerr);
   }
 
 private:
+  rclcpp::Clock clock;
   std::mutex mutex;
   float vcc;
   size_t i2cerr;
   size_t i2cerr_last;
+  rclcpp::Time last_rcd;
 };
 
 
@@ -419,16 +486,20 @@ public:
     mem_diag("APM Memory"),
     hwst_diag("APM Hardware"),
     sys_diag("System"),
-    batt_diag("Battery"),
     conn_heartbeat_mav_type(MAV_TYPE::ONBOARD_CONTROLLER),
     version_retries(RETRIES_COUNT),
     disable_diag(false),
-    has_battery_status(false),
-    battery_voltage(0.0)
+    has_battery_status0(false)
   {
+    batt_diag.reserve(MAX_NR_BATTERY_STATUS);
+    batt_diag.emplace_back("Battery");
+    for (size_t i = 2; i <= MAX_NR_BATTERY_STATUS; ++i) {
+      batt_diag.emplace_back(utils::format("Battery %u", i));
+    }
+
     enable_node_watch_parameters();
 
-    node_declate_and_watch_parameter(
+    node_declare_and_watch_parameter(
       "conn_timeout", 10.0, [&](const rclcpp::Parameter & p) {
         auto conn_timeout = rclcpp::Duration::from_seconds(p.as_double());
 
@@ -438,35 +509,41 @@ public:
           std::bind(&SystemStatusPlugin::timeout_cb, this));
       });
 
-    node_declate_and_watch_parameter(
-      "min_voltage", 10.0, [&](const rclcpp::Parameter & p) {
-        auto min_voltage = p.as_double();
-
-        batt_diag.set_min_voltage(min_voltage);
+    node_declare_and_watch_parameter(
+      "min_voltage", std::vector<double>({10.0}), [&](const rclcpp::Parameter & p) {
+        min_voltage = p.as_double_array();
+        for (size_t i = 0; i < batt_diag.size() && i < min_voltage.size(); ++i) {
+          batt_diag[i].set_min_voltage(min_voltage[i]);
+        }
       });
 
-    node_declate_and_watch_parameter(
+    node_declare_and_watch_parameter(
       "disable_diag", false, [&](const rclcpp::Parameter & p) {
         disable_diag = p.as_bool();
 
         if (!disable_diag) {
           uas->diagnostic_updater.add(sys_diag);
-          uas->diagnostic_updater.add(batt_diag);
+          for (size_t i = 0; i < batt_diag.size() && i < min_voltage.size(); ++i) {
+            uas->diagnostic_updater.add(batt_diag[i]);
+          }
+
         } else {
           uas->diagnostic_updater.removeByName(sys_diag.getName());
-          uas->diagnostic_updater.removeByName(batt_diag.getName());
           uas->diagnostic_updater.removeByName(mem_diag.getName());
           uas->diagnostic_updater.removeByName(hwst_diag.getName());
+          for (auto & d : batt_diag) {
+            uas->diagnostic_updater.removeByName(d.getName());
+          }
         }
       });
 
-    node_declate_and_watch_parameter(
+    node_declare_and_watch_parameter(
       "heartbeat_mav_type", utils::enum_to_name(
         conn_heartbeat_mav_type), [&](const rclcpp::Parameter & p) {
         conn_heartbeat_mav_type = utils::mav_type_from_str(p.as_string());
       });
 
-    node_declate_and_watch_parameter(
+    node_declare_and_watch_parameter(
       "heartbeat_rate", 1.0, [&](const rclcpp::Parameter & p) {
         auto rate_d = p.as_double();
 
@@ -494,28 +571,23 @@ public:
     extended_state_pub = node->create_publisher<mavros_msgs::msg::ExtendedState>(
       "extended_state", state_qos);
     estimator_status_pub = node->create_publisher<mavros_msgs::msg::EstimatorStatus>(
-      "estimator_status", state_qos
-    );
+      "estimator_status", state_qos);
     batt_pub = node->create_publisher<BatteryMsg>("battery", sensor_qos);
 
     statustext_pub = node->create_publisher<mavros_msgs::msg::StatusText>(
-      "statustext/recv",
-      sensor_qos);
+      "statustext/recv", sensor_qos);
     statustext_sub = node->create_subscription<mavros_msgs::msg::StatusText>(
-      "statustext/send",
-      sensor_qos,
+      "statustext/send", sensor_qos,
       std::bind(&SystemStatusPlugin::statustext_cb, this, _1));
 
     srv_cg = node->create_callback_group(rclcpp::CallbackGroupType::MutuallyExclusive);
 
-    mode_srv =
-      node->create_service<mavros_msgs::srv::SetMode>(
+    mode_srv = node->create_service<mavros_msgs::srv::SetMode>(
       "set_mode",
       std::bind(
         &SystemStatusPlugin::set_mode_cb, this, _1,
         _2), rmw_qos_profile_services_default, srv_cg);
-    stream_rate_srv =
-      node->create_service<mavros_msgs::srv::StreamRate>(
+    stream_rate_srv = node->create_service<mavros_msgs::srv::StreamRate>(
       "set_stream_rate",
       std::bind(
         &SystemStatusPlugin::set_rate_cb, this, _1,
@@ -532,8 +604,9 @@ public:
 
     uas->diagnostic_updater.add(hb_diag);
 
-    autopilot_version_timer =
-      node->create_wall_timer(1s, std::bind(&SystemStatusPlugin::autopilot_version_cb, this));
+    autopilot_version_timer = node->create_wall_timer(
+      1s, std::bind(&SystemStatusPlugin::autopilot_version_cb, this),
+      srv_cg);
 
     // init state topic
     publish_disconnection();
@@ -560,7 +633,7 @@ private:
   MemInfo mem_diag;
   HwStatus hwst_diag;
   SystemStatusDiag sys_diag;
-  BatteryStatusDiag batt_diag;
+  std::vector<BatteryStatusDiag> batt_diag;
 
   rclcpp::TimerBase::SharedPtr timeout_timer;
   rclcpp::TimerBase::SharedPtr heartbeat_timer;
@@ -584,8 +657,9 @@ private:
   static constexpr int RETRIES_COUNT = 6;
   int version_retries;
   bool disable_diag;
-  bool has_battery_status;
+  bool has_battery_status0;
   float battery_voltage;
+  std::vector<double> min_voltage;
 
   using M_VehicleInfo = std::unordered_map<uint16_t, mavros_msgs::msg::VehicleInfo>;
   M_VehicleInfo vehicles;
@@ -819,18 +893,24 @@ private:
     mavlink::common::msg::SYS_STATUS & stat,
     plugin::filter::SystemAndOk filter [[maybe_unused]])
   {
+    using MC = mavlink::minimal::MAV_COMPONENT;
+    if (static_cast<MC>(msg->compid) == MC::COMP_ID_GIMBAL) {
+      return;
+    }
+
     float volt = stat.voltage_battery / 1000.0f;                // mV
     float curr = stat.current_battery / 100.0f;                 // 10 mA or -1
     float rem = stat.battery_remaining / 100.0f;                // or -1
 
     battery_voltage = volt;
     sys_diag.set(stat);
-    batt_diag.set(volt, curr, rem);
 
-    if (has_battery_status) {
+    if (has_battery_status0) {
       return;
     }
 
+    batt_diag[0].set(volt, curr, rem);
+
     auto batt_msg = BatteryMsg();
     batt_msg.header.stamp = node->now();
     batt_msg.voltage = volt;
@@ -871,7 +951,7 @@ private:
     mavlink::ardupilotmega::msg::MEMINFO & mem,
     plugin::filter::SystemAndOk filter [[maybe_unused]])
   {
-    mem_diag.set(mem.freemem, mem.brkval);
+    mem_diag.set(std::max(static_cast<uint32_t>(mem.freemem), mem.freemem32), mem.brkval);
   }
 
   void handle_hwstatus(
@@ -928,8 +1008,6 @@ private:
     // PX4.
     using BT = mavlink::common::MAV_BATTERY_TYPE;
 
-    has_battery_status = true;
-
     auto batt_msg = BatteryMsg();
     batt_msg.header.stamp = node->now();
     batt_msg.voltage = battery_voltage;
@@ -979,19 +1057,61 @@ private:
     batt_msg.present = true;
 
     batt_msg.cell_voltage.clear();
-    batt_msg.cell_voltage.reserve(bs.voltages.size());
+    batt_msg.cell_voltage.reserve(bs.voltages.size() + bs.voltages_ext.size());
+    float cell_voltage;
+    float voltage_acc = 0.0f;
+    float total_voltage = 0.0f;
+    // 65,534V cell voltage means that the next element in the array must be added to this one
+    constexpr float coalesce_voltage = (UINT16_MAX - 1) * 0.001f;
     for (auto v : bs.voltages) {
       if (v == UINT16_MAX) {
         break;
       }
 
-      batt_msg.cell_voltage.push_back(v / 1000.0f);                    // 1 mV
+      if (v == UINT16_MAX - 1) {    // cell voltage is above 65,534V
+        voltage_acc += coalesce_voltage;
+        continue;                   // add to the next array element to get the correct voltage
+      }
+
+      cell_voltage = voltage_acc + (v * 0.001f);    // 1 mV
+      voltage_acc = 0.0f;
+      batt_msg.cell_voltage.push_back(cell_voltage);
+      total_voltage += cell_voltage;
     }
+    for (auto v : bs.voltages_ext) {
+      if (v == UINT16_MAX || v == 0) {
+        // this one is different from the for loop above to support mavlink2 message truncation
+        break;
+      }
+
+      if (v == UINT16_MAX - 1) {
+        // cell voltage is above 65,534V
+        // add to the next array element to get the correct voltage
+        voltage_acc += coalesce_voltage;
+        continue;
+      }
+
+      cell_voltage = voltage_acc + (v * 0.001f);    // 1 mV
+      voltage_acc = 0.0f;
+      batt_msg.cell_voltage.push_back(cell_voltage);
+      total_voltage += cell_voltage;
+    }
+    batt_msg.voltage = total_voltage;
 
     batt_msg.location = utils::format("id%u", bs.id);
     batt_msg.serial_number = "";
 
     batt_pub->publish(batt_msg);
+
+
+    if (bs.id == 0) {
+      has_battery_status0 = true;
+    }
+
+    if (!disable_diag && bs.id < MAX_NR_BATTERY_STATUS) {
+      batt_diag[bs.id].set(total_voltage, batt_msg.current, batt_msg.percentage);
+      batt_diag[bs.id].setcell_v(batt_msg.cell_voltage);
+    }
   }
 
   void handle_estimator_status(
@@ -1123,7 +1243,7 @@ private:
 
   void connection_cb(bool connected) override
   {
-    has_battery_status = false;
+    has_battery_status0 = false;
 
     // if connection changes, start delayed version request
     version_retries = RETRIES_COUNT;
